
'' Table of tokens (zero-terminated strings).  The program is scanned for these
'' and the original text is replaced by the number of the token + 128.  The use
'' of tokens simplifies the parsing of the program.  Note that a token should
'' not be a prefix of another token (RUN and RUNIT for example).  If you need
'' something like this, the longer token should come first.

dat
   tok0  byte "IF", 0
''       ...
   tok58 byte "DUMP", 0     ' DUMP <expr>,<expr> or DUMP [<expr>],<expr>
   tok59 byte "SERVO", 0    ' SERVO <expr>,<expr>
                            ' SERVO <expr>,<expr>,<expr>
   tok60 byte "BAUD", 0     ' BAUD = <expr>
   tok61 byte "REBOOT", 0   ' REBOOT

'' This is a table of offsets of the token text strings.  The token scanner
'' starts at "toks" and stops at "tokx" (or when a match occurs).  You can
'' add to the end of the table.  The actual token stored is +128.

   toks  word @tok0, @tok1, @tok2, @tok3, @tok4, @tok5, @tok6, @tok7
'        ...
         word @tok56, @tok57, @tok58, @tok59, @tok60, @tok61
   tokx  word

'' Process a possible EEPROM address, I/O pin number, or other [] bracketed
'' value depending on the "delim" value.  If ".", then a simple I/O pin
'' number is expected and checked for range 0-31.  A pin range like "1..5"
'' is allowed.  If present, the first value is returned in the high word of
'' the result and the second value is returned in the low word of the result.
'' If ",", then an EEPROM address is expected with the I/O pin number as the
'' first value, then a comma followed by the EEPROM address.  The pin number
'' in this case must be even and the EEPROM address in the range 0..$7FFFF.
'' The pin number is shifted into position for the sdspiFemto low level I2C
'' routines to use directly.  Any other "delim" value allows any single
'' expression value to be in the brackets and this is returned unchanged.

pri getAddress(delim) | t

'' This is the "lowest level" routine in the recursive descent expression
'' parser.

pri factor | tok, t, i
   tok := spaces
   tp++
   case tok
      "(":                           ' Recurse to handle ( <expr> )
         t := expr
         if spaces <> ")"
            abort @syn
         tp++
         return t
      "a".."z","A".."Z":             ' Scalar variable, return value
         return vars[fixvar(tok)]
      158: ' INA [ <expr>{..<expr>} ] ' Input register (INA) bit/bits
         t := getAddress(".")
         if t > $FFFF
           tok := t & $FF
           t := (t >> 8) & $FF
           repeat i from t to tok
              outputs &= ! |< i
           dira[t..tok]~
           return ina[t..tok]
         else
           outputs &= ! |< t
           dira[t]~
           return ina[t]
      166: ' BYTE [ <expr> ]         ' Hub memory byte value given address
         return byte[getAddress(" ")]
      167: ' WORD [ <expr> ]         ' Hub memory word value given address
         return word[getAddress(" ")]
      168: ' LONG [ <expr> ]         ' Hub memory long value given address
         return long[getAddress(" ")]
      181: ' EEPROM [ <expr> ]       ' EEPROM byte value given pin# / address
         t := getAddress(",")
         t := i2c.readByte(i2c#bootPin,i2c#EEPROM,t)
         if t < 0
            abort string("EEPROM read")
         return t
      other:
         abort @syn

'' Handle "pseudo-assignment" where some special token comes first, then an
'' assignment (equal sign) and an arbitrary expression.  Return the value of
'' the expression.

pri specialExpr

'' This routine handles all statements and sequences of statements separated
'' by ":".  It handles the interrupted multiple interpretation of PAUSEs as
'' well as the FOR / NEXT looping.

pri texec | ht, nt, restart, thisLine, uS, a,b,c,d,w, f0,f1,f2,f3,f4,f5,f6,f7
   uS := clkfreq / 1_000_000
   thisLine := tp - 2
   restart := 1
   repeat while restart
      restart := 0
      ht := spaces
      if ht == 0                     ' Quit at the end of the line
         return
      nt := skipspaces
      if isvar(ht) and nt == "="     ' Variable assignment
         tp++
         vars[fixvar(ht)] := expr
      elseif ht => 128
         case ht
            181: ' EEPROM [ <expr> ] = <expr>
               a := getAddress(",")  ' Set EEPROM byte to specified expression
               if i2c.writeByte(i2c#bootPin,i2c#EEPROM,a,specialExpr)
                  abort string("EEPROM write")
               repeat while i2c.WriteWait(i2c#bootPin,i2c#EEPROM,a)
                  if cnt - w > clkfreq / 10
                     abort string("EEPROM timeout")
            187: ' SERVO <expr>,<expr>{,<expr>}
               a := expr             ' Set servo to specified position
               if a < 0 or a > 15    ' Parameters are I/O pin #, pulse width
                  abort string("Servo pin < 0 or > 15")
               if spaces <> ","
                  abort @syn
               skipspaces
               b := expr             ' Invalid values turn off servo
               if spaces == ","
                  skipspaces
                  c := expr          ' Optional is ramp speed value
                  if c < 0 or c > 63
                     abort string("Ramp speed < 0 or > 63")
                  srv.SetRamp(a,b,RampSpeed[c])
               else
                  srv.Set(a,b)
               if b < srv#LowRange or b > srv#HighRange
                  servosOn &= ! |< a ' Mark I/O pin not in use
               else
                  servosOn |= |< a   ' Mark I/O pin in use
            188: ' BAUD = <expr>     ' Set Baud and reinitialize serial I/O
               initSerial(specialExpr)
            189: ' REBOOT            ' Reboot
               reboot
            139: ' RND = <expr>      ' Initialize the random seed
               rv := specialExpr
      else                           ' Unknown statement token
         abort(@syn)
      if spaces == ":"               ' ";" separates statements on a line
         restart := 1
         tp++
