
'' Main method.  Initializes program, displays banner, then does main loop
'' which reads an input line, processes it, and displays any error messages.

PUB main | err, s

'' Displays error message (address of string in "err") including line number
'' if program is executing.  Sets next line address to end of program to stop
'' execution.

PRI showError(err)

'' Get a line of text from the input device.  Handles backspace and return.
'' If input is programming port, echo it and handle backspace.  If input is
'' auxiliary serial port, flush transmitter and receiver initially and don't
'' echo.  Set "tp" to start of text string.

PRI getline | i, c

'' Display a tokenized text line.  Replace any tokens (byte >= 128) with the
'' text from the token table.  If an unknown token is found, display its value
'' in braces {}.  If token is not REM, put a space after it.

pri putlinet(s) | c, ntoks

'' Skip spaces in the input line ("tp" is the pointer) and return the first
'' non-blank character found.  A zero byte also terminates the operation.

pri spaces | c

'' Skip a single character in the input line (unless the zero byte terminator)
'' and call "spaces" to return the next non-blank.

pri skipspaces

'' The next input item is assumed to be a decimal number.  Return the value of
'' that number and leave "tp" pointing to the non-digit following the number.
'' If the next input character is not a digit, return zero.

pri parseliteral | r, c

'' Move the remainder of the program beginning at "at" upwards in memory a
'' distance "delta" to make room for a new line.

pri movprog(at, delta)

'' Convert upper case letters to lower case and return the index of the letter

pri fixvar(c)

'' Return true if the supplied character is a letter (a variable name)

pri isvar(c)

'' Scan the input line.  Replace any tokens with their 128+ code.  Any
'' characters within double quotes (") are copied literally as are comments
'' beginning with the REM token.

pri tokenize | tok, c, at, put, state, i, j, ntoks

'' Return the byte-aligned word value at location "loc"

pri wordat(loc)

'' Return the address of the line in the program that has a line number greater
'' or equal to the supplied line number.

pri findline(lineno) | at

'' Scan the line number, tokenize the rest of the line, and insert it
'' in the appropriate position in the stored program.  Delete any existing
'' line with the same line number.

pri insertline | lineno, fc, loc, locat, newlen, oldlen

'' Clear all variables to zero, clear the data stack, initialize the PAUSE
'' time, and set the program execution point to the beginning of the program

pri clearvars

'' Clear the current program and the data stack

pri newprog

'' Clear the program space, variables, and data stack

pri clearall

'' Check for a possible data stack overflow and push the current program
'' pointer into the stack (for a GOSUB)

pri pushstack

'' Process a possible EEPROM address, I/O pin number, or other [] bracketed
'' value depending on the "delim" value.  If ".", then a simple I/O pin
'' number is expected and checked for range 0-31.  A pin range like "1..5"
'' is allowed.  If present, the first value is returned in the high word of
'' the result and the second value is returned in the low word of the result.
'' If ",", then an EEPROM address is expected with the I/O pin number as the
'' first value, then a comma followed by the EEPROM address.  The pin number
'' in this case must be even and the EEPROM address in the range 0..$7FFFF.
'' The pin number is shifted into position for the sdspiFemto low level I2C
'' routines to use directly.  Any other "delim" value allows any single
'' expression value to be in the brackets and this is returned unchanged.

pri getAddress(delim) | t

'' This is the "lowest level" routine in the recursive descent expression
'' parser.

pri factor | tok, t, i

pri shifts | tok, t                  ' Shift operations

pri bitFactor | tok, t

pri bitTerm | tok, t

pri term | tok, t

pri arithExpr | tok, t

pri compare | op, a, b, c            ' Comparison operations

pri logicNot | tok

pri logicAnd | t, tok

pri expr | tok, t

'' Handle "pseudo-assignment" where some special token comes first, then an
'' assignment (equal sign) and an arbitrary expression.  Return the value of
'' the expression.

pri specialExpr

'' This routine handles all statements and sequences of statements separated
'' by ":".  It handles the interrupted multiple interpretation of PAUSEs as
'' well as the FOR / NEXT looping.

pri texec | ht, nt, restart, thisLine, uS, a,b,c,d,w, f0,f1,f2,f3,f4,f5,f6,f7

'' Reinitialize the serial port

pri initSerial(Baud)

'' Process a line of program text.  This can be a string or a user entered line.
'' Check for a break key and halt execution if present.  Proceed from statement
'' to statement if not stopped.  Clean up servos, I/O state, and PAUSE bookkeeping
'' if program stopped.

pri doline(s) | c                    ' Execute the string in s or wait for input

'' Return the value of a numeric constant next in the input stream.
'' This can be a quoted character constant or a decimal, hexadecimal,
'' or binary format numeric constant.

PRI getAnyNumber | c, t

'' Convert hexadecimal character to the corresponding value or -1 if invalid.

PRI hexDigit(c)

'' This routine dumps a portion of the RAM/ROM to the display (pin == -1).
'' If pin is not -1, it is an EEPROM address in the form required by the
'' I2C routines in i2cSpiInit.  The specified address is or'd with this.
'' The format is 8 bytes wide with hexadecimal and ASCII.

PUB dumpMemory(pin,addr,size) | i, c, p, first, buf0, buf1, buf2
